import torch
from emu_base import compute_noise_from_lindbladians, matmul_2x2_with_batched
from emu_sv.density_matrix_state import DensityMatrix

dtype = torch.complex128
sigmax = torch.tensor([[0.0, 1.0], [1.0, 0.0]], dtype=dtype)
sigmay = torch.tensor([[0.0, -1.0j], [1.0j, 0.0]], dtype=dtype)
n_op = torch.tensor([[0.0, 0.0], [0.0, 1.0]], dtype=dtype)


class RydbergLindbladian:
    """
    Apply the Lindblad superoperator ℒ to a density matrix 𝜌,  ℒ(𝜌).

    This class implements
    H @𝜌- H @ 𝜌 + i ∑ₖ − 1/2 Aₖ† Aₖ 𝜌 − 1/2 𝜌 Aₖ^† Aₖ + Aₖ 𝜌 Aₖ^†,
    where A_k is a jump operator and H is the Rydberg Hamiltonian.
    The complex -𝑖, will be multiplied in the evolution.

    Only works with effective noise channels, i.e., the jump or collapse
    operators. For more information, see:
    https://pulser.readthedocs.io/en/stable/tutorials/effective_noise.html

    Attributes:
        nqubits (int): number of qubits in the system.
        omegas (torch.Tensor): amplited frequencies  Ωⱼ for each qubit, divided by 2.
        deltas (torch.Tensor): detunings 𝛿ᵢ for each qubit.
        phis (torch.Tensor): phases 𝜙ᵢ for each qubit.
        interaction_matrix (torch.Tensor): interaction_matrix (torch.Tensor): matrix Uᵢⱼ
            representing pairwise Rydberg interaction strengths between qubits.
        pulser_linblads (list[torch.Tensor]): List of 2x2 local Lindblad (jump)
            operators acting on each qubit.
        device (torch.device): device on which tensors are allocated. cpu or gpu: cuda.
        complex (bool): flag indicating whether any drive phase is nonzero
            (i.e., complex Hamiltonian terms).
        diag (torch.Tensor): precomputed diagonal interaction term for the density matrix evolution.

    Methods:
        apply_local_op_to_density_matrix(density_matrix, local_op, target_qubit):
            Applies a local operator to the density matrix from the left: L @ ρ.

        apply_density_matrix_to_local_op_T(density_matrix, local_op, target_qubit):
            Applies a daggered local operator to the density matrix from the right: ρ @ L†.

        __matmul__(density_matrix):
            Applies the full Lindbladian superoperator to the input density matrix,
            including coherent evolution and all dissipation channels.
    """

    def __init__(
        self,
        omegas: torch.Tensor,
        deltas: torch.Tensor,
        phis: torch.Tensor,
        pulser_linblads: list[torch.Tensor],
        interaction_matrix: torch.Tensor,
        device: torch.device,
    ):
        self.nqubits: int = len(omegas)
        self.omegas: torch.Tensor = omegas / 2.0
        self.deltas: torch.Tensor = deltas
        self.phis: torch.Tensor = phis
        self.interaction_matrix: torch.Tensor = interaction_matrix
        self.pulser_linblads: list[torch.Tensor] = pulser_linblads
        self.device: torch.device = device
        self.complex = self.phis.any()

        self.diag: torch.Tensor = self._create_diagonal()

    def _create_diagonal(self) -> torch.Tensor:
        """
        Return the diagonal elements of the Rydberg Hamiltonian matrix
        concerning only the interaction

            H.diag =  ∑ᵢ﹥ⱼUᵢⱼnᵢnⱼ
        """
        diag = torch.zeros(2**self.nqubits, dtype=dtype, device=self.device)

        for i in range(self.nqubits):
            diag = diag.view(2**i, 2, -1)
            i_fixed = diag[:, 1, :]
            for j in range(i + 1, self.nqubits):
                i_fixed = i_fixed.view(2**i, 2 ** (j - i - 1), 2, -1)
                # replacing i_j_fixed by i_fixed breaks the code :)
                i_j_fixed = i_fixed[:, :, 1, :]
                i_j_fixed += self.interaction_matrix[i, j]
        return diag.view(-1)

    def apply_local_op_to_density_matrix(
        self,
        density_matrix: torch.Tensor,
        local_op: torch.Tensor,
        target_qubit: int,
    ) -> torch.Tensor:
        """
        Calculate a local operator (2x2) L being multiplied by a density matrix ρ
        from the left
        Return L @ ρ
        """

        orignal_shape = density_matrix.shape
        density_matrix = density_matrix.view(2**target_qubit, 2, -1)
        if density_matrix.is_cpu:
            density_matrix = local_op @ density_matrix
        else:
            density_matrix = matmul_2x2_with_batched(local_op, density_matrix)

        return density_matrix.view(orignal_shape)

    def apply_density_matrix_to_local_op_T(
        self,
        density_matrix: torch.Tensor,
        local_op: torch.Tensor,
        target_qubit: int,
    ) -> torch.Tensor:
        """
        Calculates a density matrix ρ being multiplied by a daggered local (2x2)
          operator L† from the right,

        return:  ρ @L†
        """

        orignal_shape = density_matrix.shape
        density_matrix = density_matrix.view(2 ** (target_qubit + self.nqubits), 2, -1)
        if density_matrix.is_cpu:
            density_matrix = local_op.conj() @ density_matrix
        else:
            density_matrix = matmul_2x2_with_batched(local_op.conj(), density_matrix)

        return density_matrix.view(orignal_shape)

    def h_eff(
        self,
        density_matrix: torch.Tensor,
        lindblad_ops: torch.Tensor = torch.zeros(2, 2, dtype=dtype),
    ) -> torch.Tensor:
        """Compute the effective Hamiltonian, Heff = Hρ  -0.5i ∑ₖ Lₖ† Lₖ ρ, applied
        to a density matrix ρ.
        - libdlad_ops by default are 2x2 zero matrix"""
        H_den_matrix = torch.zeros_like(density_matrix, dtype=dtype, device=self.device)

        for qubit in range(len(self.omegas)):
            H_q = self._local_terms_hamiltonian(qubit, lindblad_ops.to(self.device))
            H_den_matrix += self.apply_local_op_to_density_matrix(
                density_matrix, H_q, qubit
            )

        H_den_matrix += self._apply_interaction_terms(density_matrix)
        return H_den_matrix

    def _local_terms_hamiltonian(
        self, qubit: int, lindblad_ops: torch.Tensor
    ) -> torch.Tensor:
        """Construct the Hamiltonian for single qubit terms:
        ∑ᵢ 𝛺ᵢ /2 𝜎ₓ^ i − 𝛿ⁱ nᵢ + jump operators terms , including the sum
        of  Lindblad terms, when 𝜙ᵢ equals to 0.0.
        When 𝜙ᵢ not equals to 0.0:
        ∑ᵢ 𝛺ᵢ /2 (cos (𝜙ᵢ) 𝜎ₓ^ i + sin(𝜙ᵢ)𝜎_yⁱ) − 𝛿ⁱ nᵢ + jump operators terms
        """
        omega = self.omegas[qubit]
        delta = self.deltas[qubit]

        sigma_x = sigmax.to(device=self.device)
        n = n_op.to(device=self.device)

        if not self.complex:
            return omega * sigma_x - delta * n + lindblad_ops.to(self.device)

        phi = self.phis[qubit]
        sigma_y = sigmay.to(device=self.device)
        complex_part = torch.cos(phi) * sigma_x + torch.sin(phi) * sigma_y
        return omega * complex_part - delta * n + lindblad_ops.to(self.device)

    def _apply_interaction_terms(self, density_matrix: torch.Tensor) -> torch.Tensor:
        """Apply the interaction terms ∑ᵢⱼ Uᵢⱼ nᵢ nⱼ to the density matrix."""
        return self.diag.view(-1, 1) * density_matrix

    def __matmul__(self, density_matrix: torch.Tensor) -> torch.Tensor:
        """Apply the i*RydbergLindbladian operator to the density matrix ρ
        in the following way:
        Define and effective Hamiltonian
        Heff = Hρ  -0.5i ∑ₖ Lₖ† Lₖ ρ
        Then, the Lindblad operator applying to ρ is giving by
         ℒ(𝜌) = Heff - Heff^†+i*∑ₖ Lₖ ρ Lₖ†
        """

        # compute -0.5i ∑ₖ Lₖ† Lₖ
        sum_lindblad_local = compute_noise_from_lindbladians(self.pulser_linblads).to(
            self.device
        )
        # Heff = Hρ  -0.5i ∑ₖ Lₖ† Lₖ ρ
        H_den_matrix = self.h_eff(density_matrix, sum_lindblad_local)

        # Heff - Heff^†=  [H, ρ] - 0.5i ∑ₖ Lₖ† Lₖρ - ρ 0.5i ∑ₖ Lₖ† Lₖρ
        H_den_matrix = H_den_matrix - H_den_matrix.conj().T

        # compute ∑ₖ Lₖ ρ Lₖ†, last part of the Lindblad operator
        L_den_matrix_Ldag = sum(
            self.apply_density_matrix_to_local_op_T(
                self.apply_local_op_to_density_matrix(
                    density_matrix, L.to(self.device), qubit
                ),
                L.to(self.device),
                qubit,
            )
            for qubit in range(self.nqubits)
            for L in self.pulser_linblads
        )

        return H_den_matrix + 1.0j * L_den_matrix_Ldag

    def expect(self, state: DensityMatrix) -> torch.Tensor:
        """Return the energy expectation value E=tr(H𝜌)"""
        en = (self.h_eff(state.matrix)).trace()

        assert torch.allclose(en.imag, torch.zeros_like(en.imag), atol=1e-8)
        return en.real



import torch

from pulser.backend import (
    CorrelationMatrix,
    EmulationConfig,
    EnergySecondMoment,
    EnergyVariance,
    Occupation,
)
from emu_sv.density_matrix_state import DensityMatrix
from emu_sv.state_vector import StateVector
from emu_sv.dense_operator import DenseOperator
from emu_sv.hamiltonian import RydbergHamiltonian
from emu_sv.lindblad_operator import RydbergLindbladian

dtype = torch.float64


def qubit_occupation_sv_impl(
    self: Occupation,
    *,
    config: EmulationConfig,
    state: StateVector,
    hamiltonian: DenseOperator,
) -> torch.Tensor:
    """
    Custom implementation of the occupation ❬ψ|nᵢ|ψ❭ for the state vector solver.
    """
    nqubits = state.n_qudits
    occupation = torch.zeros(nqubits, dtype=dtype, device=state.vector.device)
    for i in range(nqubits):
        state_tensor = state.vector.view(2**i, 2, -1)
        # nᵢ is a projector and therefore nᵢ == nᵢnᵢ
        # ❬ψ|nᵢ|ψ❭ == ❬ψ|nᵢnᵢ|ψ❭ == ❬ψ|nᵢ * nᵢ|ψ❭ == ❬ϕ|ϕ❭ == |ϕ|**2
        occupation[i] = torch.linalg.vector_norm(state_tensor[:, 1]) ** 2
    return occupation.cpu()


def qubit_occupation_sv_den_mat_impl(
    self: Occupation,
    *,
    config: EmulationConfig,
    state: DensityMatrix,
    hamiltonian: DenseOperator,
) -> torch.Tensor:
    """
    Custom implementation of the occupation nᵢ observable for density matrix.
    The observable nᵢ is given by: I ⊗ ... ⊗  nᵢ ⊗ ...  ⊗I
    where nᵢ is the occupation operator for qubit i.
    The expectation value is given by: <nᵢ> = Tr(ρ nᵢ).

    The output will be a tensor of size (nqubits,), where each element will be the
    expectation value of the occupation operator for each qubit.
    In case of 3 atoms, the output will be a tensor of size (3,), where each element
    will be <nᵢ> = Tr(ρnᵢ), or [ <n₁>, <n₂>, <n₃> ].
    """
    nqubits = state.n_qudits
    occupation = torch.zeros(nqubits, dtype=dtype, device=state.matrix.device)
    diag_state_tensor = state.matrix.diagonal()
    for i in range(nqubits):
        state_tensor = diag_state_tensor.view(2**i, 2, 2 ** (nqubits - i - 1))[:, 1, :]
        occupation[i] = state_tensor.sum().real
    return occupation.cpu()


def correlation_matrix_sv_impl(
    self: CorrelationMatrix,
    *,
    config: EmulationConfig,
    state: StateVector,
    hamiltonian: DenseOperator,
) -> torch.Tensor:
    """
    Custom implementation of the density-density correlation ❬ψ|nᵢnⱼ|ψ❭
      for the state vector solver.
    TODO: extend to arbitrary two-point correlation ❬ψ|AᵢBⱼ|ψ❭
    """
    nqubits = state.n_qudits
    correlation = torch.zeros(nqubits, nqubits, dtype=dtype, device=state.vector.device)

    for i in range(nqubits):
        select_i = state.vector.view(2**i, 2, -1)
        select_i = select_i[:, 1]
        correlation[i, i] = torch.linalg.vector_norm(select_i) ** 2
        for j in range(i + 1, nqubits):  # select the upper triangle
            select_i = select_i.view(2**i, 2 ** (j - i - 1), 2, -1)
            select_ij = select_i[:, :, 1, :]
            correlation[i, j] = torch.linalg.vector_norm(select_ij) ** 2
            correlation[j, i] = correlation[i, j]

    return correlation.cpu()


def correlation_matrix_sv_den_mat_impl(
    self: CorrelationMatrix,
    *,
    config: EmulationConfig,
    state: DensityMatrix,
    hamiltonian: DenseOperator,
) -> torch.Tensor:
    """
    Custom implementation of the density-density correlation <nᵢnⱼ> = Tr(ρ nᵢnⱼ)
    in the case of Lindblad noise
    """
    nqubits = state.n_qudits
    correlation = torch.zeros(nqubits, nqubits, dtype=dtype)
    state_diag_matrix = state.matrix.diagonal()
    for i in range(nqubits):  # applying ni
        shapei = (2**i, 2, 2 ** (nqubits - i - 1))
        state_diag_ni = state_diag_matrix.view(*shapei)[:, 1, :]
        correlation[i, i] = state_diag_ni.sum().real  # diagonal
        for j in range(i + 1, nqubits):
            shapeij = (2**i, 2 ** (j - i - 1), 2, 2 ** (nqubits - 1 - j))
            state_diag_ni_nj = state_diag_ni.view(*shapeij)[:, :, 1, :]
            correlation[i, j] = state_diag_ni_nj.sum().real
            correlation[j, i] = correlation[i, j]
    return correlation.cpu()


def energy_variance_sv_impl(
    self: EnergyVariance,
    *,
    config: EmulationConfig,
    state: StateVector,
    hamiltonian: RydbergHamiltonian,
) -> torch.Tensor:
    """
    Custom implementation of the energy variance ❬ψ|H²|ψ❭-❬ψ|H|ψ❭² for the state vector solver.
    """
    hstate = hamiltonian * state.vector
    h_squared = torch.vdot(hstate, hstate).real
    energy = torch.vdot(state.vector, hstate).real
    en_var: torch.Tensor = h_squared - energy**2
    return en_var.cpu()


def energy_variance_sv_den_mat_impl(
    self: EnergyVariance,
    *,
    config: EmulationConfig,
    state: DensityMatrix,
    hamiltonian: RydbergLindbladian,
) -> torch.Tensor:
    """
    Custom implementation of the energy variance tr(ρH²)-tr(ρH)² for the
    lindblad equation solver.
    """
    h_dense_matrix = hamiltonian.h_eff(state.matrix)  # Hρ
    gpu = state.matrix.is_cuda
    h_squared_dense_mat = hamiltonian.expect(
        DensityMatrix(h_dense_matrix, gpu=gpu)
    )  # tr(ρH²)
    en_var: torch.Tensor = h_squared_dense_mat - hamiltonian.expect(state) ** 2  # tr(ρH)²
    return en_var.cpu()


def energy_second_moment_sv_impl(
    self: EnergySecondMoment,
    *,
    config: EmulationConfig,
    state: StateVector,
    hamiltonian: RydbergHamiltonian,
) -> torch.Tensor:
    """
    Custom implementation of the second moment of energy ❬ψ|H²|ψ❭
    for the state vector solver.
    """
    hstate = hamiltonian * state.vector
    en_2_mom: torch.Tensor = torch.vdot(hstate, hstate).real
    return en_2_mom.cpu()


def energy_second_moment_den_mat_impl(
    self: EnergyVariance,
    *,
    config: EmulationConfig,
    state: DensityMatrix,
    hamiltonian: RydbergLindbladian,
) -> torch.Tensor:
    """
    Custom implementation of the second moment of energy tr(ρH²) for the
    lindblad equation solver.
    """
    h_dense_matrix = hamiltonian.h_eff(state.matrix)  # Hρ
    gpu = state.matrix.is_cuda

    return hamiltonian.expect(
        DensityMatrix(h_dense_matrix, gpu=gpu)
    ).cpu()  # tr(ρH²) = tr(ρ H H)



from pulser.backend import EmulatorBackend
from pulser.backend import Results
from emu_sv.sv_config import SVConfig
from emu_sv.sv_backend_impl import create_impl


class SVBackend(EmulatorBackend):
    """
    A backend for emulating Pulser sequences using state vectors and sparse matrices.
    Noisy simulation is supported by solving the Lindblad equation and using effective
    noise channel or jump operators
    """

    default_config = SVConfig()

    def run(self) -> Results:
        """
        Emulates the given sequence.

        Returns:
            the simulation results
        """
        assert isinstance(self._config, SVConfig)

        impl = create_impl(self._sequence, self._config)
        return impl._run()



from typing import Callable

from pyparsing import Any


def index_to_bitstring(nqubits: int, index: int) -> str:
    """
    Convert an integer index into its corresponding bitstring representation.
    """

    msg = f"index {index} can not exceed Hilbert space size d**{nqubits}"
    assert index < 2**nqubits, msg
    return format(index, f"0{nqubits}b")


def choose(
    state_vector_version: Callable,
    density_matrix_version: Callable,
) -> Callable:
    """Returns the observable result function that chooses the correct
    implementation based on the type of state (StateVector or DensityMatrix).
    """
    from emu_sv.state_vector import StateVector
    from emu_sv.density_matrix_state import DensityMatrix

    def result(self: Any, *, state: StateVector | DensityMatrix, **kwargs: Any) -> Any:
        if isinstance(state, StateVector):
            return state_vector_version(self, state=state, **kwargs)
        elif isinstance(state, DensityMatrix):
            return density_matrix_version(self, state=state, **kwargs)
        else:
            raise TypeError(f"Unsupported state: {type(state).__name__}")

    return result



from abc import abstractmethod
import time
import typing
import torch

from emu_sv.hamiltonian import RydbergHamiltonian
from emu_sv.lindblad_operator import RydbergLindbladian
from pulser import Sequence

from pulser.backend import Results, Observable, State, EmulationConfig
from emu_base import PulserData, get_max_rss

from emu_sv.state_vector import StateVector
from emu_sv.density_matrix_state import DensityMatrix
from emu_sv.sv_config import SVConfig
from emu_sv.time_evolution import EvolveStateVector, EvolveDensityMatrix

_TIME_CONVERSION_COEFF = 0.001  # Omega and delta are given in rad/μs, dt in ns


class Statistics(Observable):
    def __init__(
        self,
        evaluation_times: typing.Sequence[float] | None,
        data: list[float],
        timestep_count: int,
    ):
        super().__init__(evaluation_times=evaluation_times)
        self.data = data
        self.timestep_count = timestep_count

    @property
    def _base_tag(self) -> str:
        return "statistics"

    def apply(
        self,
        *,
        config: EmulationConfig,
        state: State,
        **kwargs: typing.Any,
    ) -> dict:
        """Calculates the observable to store in the Results."""
        assert isinstance(state, StateVector | DensityMatrix)
        assert isinstance(config, SVConfig)
        duration = self.data[-1]
        max_mem = get_max_rss(
            isinstance(state, StateVector)
            and state.vector.is_cuda
            or isinstance(state, DensityMatrix)
            and state.matrix.is_cuda
        )

        config.logger.info(
            f"step = {len(self.data)}/{self.timestep_count}, "
            + f"RSS = {max_mem:.3f} MB, "
            + f"Δt = {duration:.3f} s"
        )

        return {
            "RSS": max_mem,
            "duration": duration,
        }


class BaseSVBackendImpl:
    """
    This class is used to handle the state vector and density matrix evolution.
    """

    well_prepared_qubits_filter: typing.Optional[torch.Tensor]

    def __init__(self, config: SVConfig, pulser_data: PulserData):
        self._config = config
        self._pulser_data = pulser_data
        self.target_times = pulser_data.target_times
        self.omega = pulser_data.omega
        self.delta = pulser_data.delta
        self.phi = pulser_data.phi
        self.nsteps = pulser_data.omega.shape[0]
        self.nqubits = pulser_data.omega.shape[1]
        self.full_interaction_matrix = pulser_data.full_interaction_matrix
        self.state: State
        self.time = time.time()
        self.results = Results(atom_order=(), total_duration=self.target_times[-1])
        self.statistics = Statistics(
            evaluation_times=[t / self.target_times[-1] for t in self.target_times],
            data=[],
            timestep_count=self.nsteps,
        )
        self._current_H: None | RydbergLindbladian | RydbergHamiltonian = None
        if self._config.initial_state is not None and (
            self._config.initial_state.n_qudits != self.nqubits
        ):
            raise ValueError(
                "Mismatch in number of atoms: initial state has "
                f"{self._config.initial_state.n_qudits} and the sequence has {self.nqubits}"
            )
        self.init_dark_qubits()

        if (
            self._config.initial_state is not None
            and self._config.noise_model.state_prep_error > 0.0
        ):
            raise NotImplementedError(
                "Initial state and state preparation error can not be together."
            )

    def init_dark_qubits(self) -> None:
        if self._config.noise_model.state_prep_error > 0.0:
            bad_atoms = self._pulser_data.hamiltonian.bad_atoms
            self.well_prepared_qubits_filter = torch.tensor(
                [bool(bad_atoms[x]) for x in self._pulser_data.qubit_ids]
            )
        else:
            self.well_prepared_qubits_filter = None

        if self.well_prepared_qubits_filter is not None:

            self.full_interaction_matrix[self.well_prepared_qubits_filter, :] = 0.0
            self.full_interaction_matrix[:, self.well_prepared_qubits_filter] = 0.0
            self.omega[:, self.well_prepared_qubits_filter] = 0.0
            self.delta[:, self.well_prepared_qubits_filter] = 0.0
            self.phi[:, self.well_prepared_qubits_filter] = 0.0

    def step(self, step_idx: int) -> None:
        """One step of the evolution"""
        dt = self._compute_dt(step_idx)
        self._evolve_step(dt, step_idx)
        self._apply_observables(step_idx)
        self._save_statistics(step_idx)

    def _compute_dt(self, step_idx: int) -> float:
        return self.target_times[step_idx + 1] - self.target_times[step_idx]

    @abstractmethod
    def _evolve_step(self, dt: float, step_idx: int) -> None:
        """One step evolution"""

    def _apply_observables(self, step_idx: int) -> None:
        norm_time = self.target_times[step_idx + 1] / self.target_times[-1]
        for callback in self._config.observables:
            callback(
                self._config,
                norm_time,
                self.state,
                self._current_H,  # type: ignore[arg-type]
                self.results,
            )

    def _save_statistics(self, step_idx: int) -> None:
        norm_time = self.target_times[step_idx + 1] / self.target_times[-1]
        self.statistics.data.append(time.time() - self.time)
        self.statistics(
            self._config,
            norm_time,
            self.state,
            self._current_H,  # type: ignore[arg-type]
            self.results,
        )
        self.time = time.time()
        self._current_H = None

    def _run(self) -> Results:
        for step in range(self.nsteps):
            self.step(step)

        return self.results


class SVBackendImpl(BaseSVBackendImpl):

    def __init__(self, config: SVConfig, pulser_data: PulserData):
        """
        For running sequences without noise. The state will evolve according
        to e^(-iH t)

        Args:
            config: The configuration for the emulator.
            pulser_data: The data for the sequence to be emulated.
        """
        super().__init__(config, pulser_data)

        self.state: StateVector = (
            StateVector.make(self.nqubits, gpu=self._config.gpu)
            if self._config.initial_state is None
            else StateVector(
                self._config.initial_state.vector.clone(),
                gpu=self._config.gpu,
            )
        )

        self.stepper = EvolveStateVector.apply

    def _evolve_step(self, dt: float, step_idx: int) -> None:
        self.state.vector, self._current_H = self.stepper(
            dt * _TIME_CONVERSION_COEFF,
            self.omega[step_idx],
            self.delta[step_idx],
            self.phi[step_idx],
            self.full_interaction_matrix,
            self.state.vector,
            self._config.krylov_tolerance,
        )


class NoisySVBackendImpl(BaseSVBackendImpl):

    def __init__(self, config: SVConfig, pulser_data: PulserData):
        """
        Initializes the NoisySVBackendImpl, master equation version.
        This class handles the Lindblad operators and
        solves the Lindblad master equation

        Args:
            config: The configuration for the emulator.
            pulser_data: The data for the sequence to be emulated.
        """

        super().__init__(config, pulser_data)

        self.pulser_lindblads = pulser_data.lindblad_ops

        self.state: DensityMatrix = (
            DensityMatrix.make(self.nqubits, gpu=self._config.gpu)
            if self._config.initial_state is None
            else DensityMatrix(
                self._config.initial_state.matrix.clone(), gpu=self._config.gpu
            )
        )

    def _evolve_step(self, dt: float, step_idx: int) -> None:
        self.state.matrix, self._current_H = EvolveDensityMatrix.evolve(
            dt * _TIME_CONVERSION_COEFF,
            self.omega[step_idx],
            self.delta[step_idx],
            self.phi[step_idx],
            self.full_interaction_matrix,
            self.state.matrix,
            self._config.krylov_tolerance,
            self.pulser_lindblads,
        )


def create_impl(sequence: Sequence, config: SVConfig) -> BaseSVBackendImpl:
    """
    Creates the backend implementation for the given sequence and config.

    Args:
        sequence: The sequence to be emulated.
        config: configuration for the emulator.

    Returns:
        An instance of SVBackendImpl.
    """
    pulse_data = PulserData(sequence=sequence, config=config, dt=config.dt)
    if pulse_data.has_lindblad_noise:
        return NoisySVBackendImpl(config, pulse_data)
    else:
        return SVBackendImpl(config, pulse_data)



from __future__ import annotations

import math
from collections import Counter
from typing import Sequence, Type, TypeVar, Mapping

import torch

from emu_sv.utils import index_to_bitstring

from emu_base import DEVICE_COUNT, apply_measurement_errors
from pulser.backend import State
from pulser.backend.state import Eigenstate

StateVectorType = TypeVar("StateVectorType", bound="StateVector")
# Default tensor data type
dtype = torch.complex128


class StateVector(State[complex, torch.Tensor]):
    """
    Represents a quantum state vector in a computational basis.

    This class extends the `State` class to handle state vectors,
    providing various utilities for initialization, normalization,
    manipulation, and measurement. The state vector must have a length
    that is a power of 2, representing 2ⁿ basis states for n qubits.

    Attributes:
        vector: 1D tensor representation of a state vector.
        gpu: store the vector on GPU if True, otherwise on CPU
    """

    def __init__(
        self,
        vector: torch.Tensor,
        *,
        gpu: bool = True,
        eigenstates: Sequence[Eigenstate] = ("r", "g"),
    ):
        # NOTE: this accepts also zero vectors.

        assert math.log2(
            len(vector)
        ).is_integer(), "The number of elements in the vector should be power of 2"

        super().__init__(eigenstates=eigenstates)
        device = "cuda" if gpu and DEVICE_COUNT > 0 else "cpu"
        self.vector = vector.to(dtype=dtype, device=device)

    @property
    def n_qudits(self) -> int:
        """The number of qudits in the state."""
        nqudits = math.log2(self.vector.view(-1).shape[0])
        return int(nqudits)

    def _normalize(self) -> None:
        """Normalizes the state vector to ensure it has unit norm.

        If the vector norm is not 1, the method scales the vector
        to enforce normalization.

        Note:
            This method is intended to be used in callbacks.
        """

        norm = torch.linalg.vector_norm(self.vector)

        if not torch.allclose(norm, torch.ones_like(norm)):
            self.vector = self.vector / norm

    @classmethod
    def zero(
        cls,
        num_sites: int,
        gpu: bool = True,
        eigenstates: Sequence[Eigenstate] = ("r", "g"),
    ) -> StateVector:
        """
        Returns a zero uninitialized "state" vector. Warning, this has no physical meaning as-is!

        Args:
            num_sites: the number of qubits
            gpu: whether gpu or cpu

        Returns:
            The zero state

        Examples:
            >>> StateVector.zero(2,gpu=False)
            tensor([0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j], dtype=torch.complex128)
        """

        device = "cuda" if gpu and DEVICE_COUNT > 0 else "cpu"
        vector = torch.zeros(2**num_sites, dtype=dtype, device=device)
        return cls(vector, gpu=gpu, eigenstates=eigenstates)

    @classmethod
    def make(cls, num_sites: int, gpu: bool = True) -> StateVector:
        """
        Returns a State vector in the ground state |00..0>.

        Args:
            num_sites: the number of qubits
            gpu: whether gpu or cpu

        Returns:
            The described state

        Examples:
            >>> StateVector.make(2,gpu=False)
            tensor([1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j], dtype=torch.complex128)


        """

        result = cls.zero(num_sites=num_sites, gpu=gpu)
        result.vector[0] = 1.0
        return result

    def inner(self, other: State) -> torch.Tensor:
        """
        Compute <self|other>. The type of other must be StateVector.

        Args:
            other: the other state

        Returns:
            the inner product
        """
        assert isinstance(other, StateVector), "Other state must be a StateVector"
        assert (
            self.vector.shape == other.vector.shape
        ), "States do not have the same shape"

        # by our internal convention inner and norm return to cpu
        return torch.vdot(self.vector, other.vector.to(self.vector.device)).cpu()

    def sample(
        self,
        *,
        num_shots: int = 1000,
        one_state: Eigenstate | None = None,
        p_false_pos: float = 0.0,
        p_false_neg: float = 0.0,
    ) -> Counter[str]:
        """
        Samples bitstrings, taking into account the specified error rates.

        Args:
            num_shots: how many bitstrings to sample
            p_false_pos: the rate at which a 0 is read as a 1
            p_false_neg: teh rate at which a 1 is read as a 0

        Returns:
            the measured bitstrings, by count
        """

        probabilities = torch.abs(self.vector) ** 2

        outcomes = torch.multinomial(probabilities, num_shots, replacement=True)

        # Convert outcomes to bitstrings and count occurrences
        counts = Counter(
            [index_to_bitstring(self.n_qudits, outcome) for outcome in outcomes]
        )

        if p_false_neg > 0 or p_false_pos > 0:
            counts = apply_measurement_errors(
                counts,
                p_false_pos=p_false_pos,
                p_false_neg=p_false_neg,
            )
        return counts

    def __add__(self, other: State) -> StateVector:
        """Sum of two state vectors

        Args:
            other: the vector to add to this vector

        Returns:
            The summed state
        """
        assert isinstance(other, StateVector), "`Other` state can only be a StateVector"
        assert (
            self.eigenstates == other.eigenstates
        ), f"`Other` state has basis {other.eigenstates} != {self.eigenstates}"
        return StateVector(
            self.vector + other.vector,
            gpu=self.vector.is_cuda,
            eigenstates=self.eigenstates,
        )

    def __rmul__(self, scalar: complex) -> StateVector:
        """Scalar multiplication

        Args:
            scalar: the scalar to multiply with

        Returns:
            The scaled state
        """
        return StateVector(
            scalar * self.vector,
            gpu=self.vector.is_cuda,
            eigenstates=self.eigenstates,
        )

    def norm(self) -> torch.Tensor:
        """Returns the norm of the state

        Returns:
            the norm of the state
        """
        nrm: torch.Tensor = torch.linalg.vector_norm(self.vector).cpu()
        return nrm

    def __repr__(self) -> str:
        return repr(self.vector)

    @classmethod
    def _from_state_amplitudes(
        cls: Type[StateVectorType],
        *,
        eigenstates: Sequence[Eigenstate],
        n_qudits: int,
        amplitudes: Mapping[str, complex],
    ) -> tuple[StateVector, Mapping[str, complex]]:
        """Transforms a state given by a string into a state vector.

        Construct a state from the pulser abstract representation
        https://pulser.readthedocs.io/en/stable/conventions.html

        Args:
            eigenstates: A tuple containing the basis states (e.g., ('r', 'g')).
            amplitudes: A dictionary mapping state strings to complex or floats amplitudes.

        Returns:
            The normalised resulting state.

        Examples:
            >>> basis = ("r","g")
            >>> n = 2
            >>> st=StateVector.from_state_string(basis=basis,
                ... nqubits=n,strings={"rr":1.0,"gg":1.0},gpu=False)
            >>> st = StateVector.from_state_amplitudes(
            ...     eigenstates=basis,
            ...     amplitudes={"rr": 1.0, "gg": 1.0}
            ... )
            >>> print(st)
            tensor([0.7071+0.j, 0.0000+0.j, 0.0000+0.j, 0.7071+0.j],
                   dtype=torch.complex128)
        """
        basis = set(eigenstates)
        if basis == {"r", "g"}:
            one = "r"
        elif basis == {"0", "1"}:
            raise NotImplementedError(
                "{'0','1'} basis is related to XY Hamiltonian, which is not implemented"
            )
        else:
            raise ValueError("Unsupported basis provided")

        accum_state = StateVector.zero(num_sites=n_qudits, eigenstates=eigenstates)

        for state, amplitude in amplitudes.items():
            bin_to_int = int(
                state.replace(one, "1").replace("g", "0"), 2
            )  # "0" basis is already in "0"
            accum_state.vector[bin_to_int] = torch.tensor([amplitude])

        accum_state._normalize()

        return accum_state, amplitudes

    def overlap(self, other: StateVector, /) -> torch.Tensor:
        return self.inner(other)


def inner(left: StateVector, right: StateVector) -> torch.Tensor:
    """
    Wrapper around StateVector.inner.

    Args:
        left:  StateVector argument
        right: StateVector argument

    Returns:
        the inner product

    Examples:
        >>> factor = math.sqrt(2.0)
        >>> basis = ("r","g")
        >>> string_state1 = {"gg":1.0,"rr":1.0}
        >>> state1 = StateVector.from_state_string(basis=basis,
            ... nqubits=nqubits,strings=string_state1)
        >>> string_state2 = {"gr":1.0/factor,"rr":1.0/factor}
        >>> state2 = StateVector.from_state_string(basis=basis,
            ... nqubits=nqubits,strings=string_state2)

        >>> state1 = StateVector.from_state_amplitudes(eigenstates=basis,
        ...     amplitudes=string_state1)
        >>> string_state2 = {"gr":1.0/factor,"rr":1.0/factor}
        >>> state2 = StateVector.from_state_amplitudes(eigenstates=basis,
        ...     amplitudes=string_state2)
        >>> inner(state1,state2).item()
        (0.49999999144286444+0j)
    """

    assert (left.vector.shape == right.vector.shape) and (left.vector.dim() == 1), (
        "Shape of left.vector and right.vector should be",
        " the same and both need to be 1D tesnor",
    )
    return left.inner(right)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import torch
from typing import Any, no_type_check
from emu_base.math.krylov_exp import krylov_exp
from emu_base.math.double_krylov import double_krylov
from emu_sv.hamiltonian import RydbergHamiltonian
from emu_sv.lindblad_operator import RydbergLindbladian


def _apply_omega_real(
    result: torch.Tensor,
    i: int,
    inds: torch.Tensor,
    source: torch.Tensor,
    alpha: complex,
) -> None:
    """Accumulate to `result` the application of ασˣᵢ on `source`"""
    result.index_add_(i, inds, source, alpha=alpha)


def _apply_omega_complex(
    result: torch.Tensor,
    i: int,
    inds: torch.Tensor,
    source: torch.Tensor,
    alpha: complex,
) -> None:
    """Accumulate to `result` the application of ασ⁺ᵢ + α*σ⁻ᵢ on `source`"""
    result.index_add_(i, inds[0], source.select(i, 0).unsqueeze(i), alpha=alpha)
    result.index_add_(
        i,
        inds[1],
        source.select(i, 1).unsqueeze(2),
        alpha=alpha.conjugate(),
    )


class DHDOmegaSparse:
    """
    Derivative of the RydbergHamiltonian respect to Omega.
        ∂H/∂Ωₖ = 0.5[cos(ϕₖ)σˣₖ + sin(ϕₖ)σʸₖ]

    If ϕₖ=0, simplifies to ∂H/∂Ωₖ = 0.5σˣₖ
    """

    def __init__(self, index: int, device: torch.device, nqubits: int, phi: torch.Tensor):
        self.index = index
        self.shape = (2**index, 2, 2 ** (nqubits - index - 1))
        self.inds = torch.tensor([1, 0], device=device)  # flips the state, for 𝜎ₓ
        self.alpha = 0.5 * torch.exp(1j * phi).item()
        if phi.is_nonzero():
            self._apply_sigmas = _apply_omega_complex
        else:  # ∂H/∂Ωₖ = 0.5σˣₖ
            self._apply_sigmas = _apply_omega_real

    def __matmul__(self, vec: torch.Tensor) -> torch.Tensor:
        vec = vec.view(vec.shape[0], *self.shape)  # add batch dimension
        result = torch.zeros_like(vec)
        self._apply_sigmas(result, 2, self.inds, vec, alpha=self.alpha)
        return result.view(vec.shape[0], -1)


class DHDPhiSparse:
    """
    Derivative of the RydbergHamiltonian respect to Phi.
        ∂H/∂ϕₖ = 0.5Ωₖ[cos(ϕₖ+π/2)σˣₖ + sin(ϕₖ+π/2)σʸₖ]
    """

    def __init__(
        self,
        index: int,
        device: torch.device,
        nqubits: int,
        omega: torch.Tensor,
        phi: torch.Tensor,
    ):
        self.index = index
        self.shape = (2**index, 2, 2 ** (nqubits - index - 1))
        self.alpha = 0.5 * (omega * torch.exp(1j * (phi + torch.pi / 2))).item()
        self.inds = torch.tensor([1, 0], device=device)  # flips the state, for 𝜎ₓ

    def __matmul__(self, vec: torch.Tensor) -> torch.Tensor:
        vec = vec.view(vec.shape[0], *self.shape)  # add batch dimension
        result = torch.zeros_like(vec)
        _apply_omega_complex(result, 2, self.inds, vec, alpha=self.alpha)
        return result.view(vec.shape[0], -1)


class DHDDeltaSparse:
    """
    Derivative of the Rydberg Hamiltonian respect to Delta:
        ∂H/∂Δᵢ = -nᵢ
    """

    def __init__(self, i: int, nqubits: int):
        self.nqubits = nqubits
        self.shape = (2**i, 2, 2 ** (nqubits - i - 1))

    def __matmul__(self, vec: torch.Tensor) -> torch.Tensor:
        result = vec.clone()
        result = result.view(vec.shape[0], *self.shape)
        result[:, :, 0] = 0.0
        return -result.view(vec.shape[0], 2**self.nqubits)


class DHDUSparse:
    """
    Derivative of the Rydberg Hamiltonian respect to the interaction matrix:
        ∂H/∂Uᵢⱼ = nᵢnⱼ
    """

    def __init__(self, i: int, j: int, nqubits: int):
        self.shape = (2**i, 2, 2 ** (j - i - 1), 2, 2 ** (nqubits - j - 1))
        self.nqubits = nqubits

    def __matmul__(self, vec: torch.Tensor) -> torch.Tensor:
        result = vec.clone()
        result = result.view(vec.shape[0], *self.shape)
        result[:, :, 0] = 0.0
        result[:, :, 1, :, 0] = 0.0
        return result.view(vec.shape[0], 2**self.nqubits)


class EvolveStateVector(torch.autograd.Function):
    """Custom autograd implementation of a step in the time evolution."""

    @staticmethod
    def evolve(
        dt: float,
        omegas: torch.Tensor,
        deltas: torch.Tensor,
        phis: torch.Tensor,
        interaction_matrix: torch.Tensor,
        state: torch.Tensor,
        krylov_tolerance: float,
    ) -> tuple[torch.Tensor, RydbergHamiltonian]:
        ham = RydbergHamiltonian(
            omegas=omegas,
            deltas=deltas,
            phis=phis,
            interaction_matrix=interaction_matrix,
            device=state.device,
        )
        op = lambda x: -1j * dt * (ham * x)
        res = krylov_exp(
            op,
            state,
            norm_tolerance=krylov_tolerance,
            exp_tolerance=krylov_tolerance,
            is_hermitian=True,
        )
        return res, ham

    @staticmethod
    def forward(
        ctx: Any,
        dt: float,
        omegas: torch.Tensor,
        deltas: torch.Tensor,
        phis: torch.Tensor,
        interaction_matrix: torch.Tensor,
        state: torch.Tensor,
        krylov_tolerance: float,
    ) -> tuple[torch.Tensor, RydbergHamiltonian]:
        """
        Returns the time evolved state
            |ψ(t+dt)〉= exp(-i dt H)|ψ(t)〉
        under the Hamiltonian H built from the input Tensor parameters, omegas, deltas, phis and
        the interaction matrix.

        Args:
            ctx (Any): context object to stash information for backward computation.
            dt (float): timestep
            omegas (torch.Tensor): 1D tensor of driving strengths for each qubit.
            deltas (torch.Tensor): 1D tensor of detuning values for each qubit.
            phis (torch.Tensor): 1D tensor of phase values for each qubit.
            interaction_matrix (torch.Tensor): matrix representing the interaction
                strengths between each pair of qubits.
            state (Tensor): input state to be evolved
            krylov_tolerance (float):
        """
        res, ham = EvolveStateVector.evolve(
            dt, omegas, deltas, phis, interaction_matrix, state, krylov_tolerance
        )
        ctx.save_for_backward(omegas, deltas, phis, interaction_matrix, state)
        ctx.dt = dt
        ctx.tolerance = krylov_tolerance
        return res, ham

    # mypy complains and I don't know why
    # backward expects same number of gradients as output of forward, gham is unused
    @no_type_check
    @staticmethod
    def backward(ctx: Any, grad_state_out: torch.Tensor, gham: None) -> tuple[
        None,
        torch.Tensor | None,
        torch.Tensor | None,
        torch.Tensor | None,
        torch.Tensor | None,
        torch.Tensor | None,
        None,
    ]:
        """
        In the backward pass we receive a Tensor containing the gradient of the loss L
        with respect to the output
            |gψ(t+dt)〉= ∂L/∂|ψ(t+dt)〉,
        and return the gradients of the loss with respect to the input tensors parameters
            - gΩⱼ = ∂L/∂Ωⱼ =〈gψ(t+dt)|dU(H,∂H/∂Ωⱼ)|ψ(t)〉
            - gΔⱼ = ∂L/∂Δⱼ =  ...
            - |gψ(t)〉= ∂L/∂|ψ(t)〉= exp(i dt H)|gψ(t+dt)〉

        Args:
            ctx (Any): context object to stash information for backward computation.
            grad_state_out (torch.Tensor): |gψ(t+dt)〉

        Return:
            grad_omegas (torch.Tensor): 1D tensor of gradients with respect to Ωⱼ for each qubit.
            grad_deltas (torch.Tensor): 1D tensor of gradients with respect to Δⱼ for each qubit.
            grad_phis (torch.Tensor): 1D tensor of gradients with respect to φⱼ for each qubit.
            grad_state_in (torch.Tensor): 1D tensor gradient with respect to the input state.

        Notes:
        Gradients are obtained by matching the total variations
            〈gψ(t+dt)|d|ψ(t+dt)〉= ∑ⱼgΔⱼ*dΔⱼ + ∑ⱼgΩⱼ*dΩⱼ + ∑ⱼgφ*dφⱼ +〈gψ(t)|d|ψ(t)〉  (1)

        For the exponential map U = exp(-i dt H), differentiating reads:
            d|ψ(t+dt)〉= dU|ψ(t)〉+ Ud|ψ(t)〉
            dU = ∑ⱼdU(H,∂H/∂Δⱼ) + ∑ⱼdU(H,∂H/∂Ωⱼ) + ∑ⱼdU(H,∂H/∂φⱼ)                      (2)

        where dU(H,E) is the Fréchet derivative of the exponential map
        along the direction E:
        - https://eprints.maths.manchester.ac.uk/1218/1/covered/MIMS_ep2008_26.pdf
        - https://en.wikipedia.org/wiki/Derivative_of_the_exponential_map

        Substituting (2) into (1) leads to the expressions of the gradients
        with respect to the input tensors above.

        Variations with respect to the Hamiltonian parameters are computed as
            gΩ = 〈gψ(t+dt)|dU(H,∂H/∂Ω)|ψ(t)〉
               = Tr( -i dt ∂H/∂Ω @ dU(H,|ψ(t)〉〈gψ(t+dt)|) ),
        where under the trace sign, ∂H/∂Ω and |ψ(t)〉〈gψ(t+dt)| can be switched.

        - The Fréchet derivative is computed in a Arnoldi-Gram-Schmidt
        decomposition in the `double_krylov` method:
            dU(H,|a〉〈b|) = Va @ dS @ Vb*
        where Va,Vb are orthogonal Krylov basis associated
        with |a〉and |b〉respectively.

        - The action of the derivatives of the Hamiltonian with
        respect to the input parameters are implemented separately in
            - ∂H/∂Ω:  `DHDOmegaSparse`
            - ∂H/∂Δ:  `DHDDeltaSparse`
            - ∂H/∂φ:  `DHDPhiSparse`
            - ∂H/∂Uᵢⱼ `DHDUSparse`

        Then, the resulting gradient respect to a generic parameter reads:
            gΩ = Tr( -i dt ∂H/∂Ω @ Vs @ dS @ Vg* )
        """
        omegas, deltas, phis, interaction_matrix, state = ctx.saved_tensors
        dt = ctx.dt
        tolerance = ctx.tolerance
        nqubits = len(omegas)

        grad_omegas, grad_deltas, grad_phis = None, None, None
        grad_int_mat = None
        grad_state_in = None

        ham = RydbergHamiltonian(
            omegas=omegas,
            deltas=deltas,
            phis=phis,
            interaction_matrix=interaction_matrix,
            device=state.device,
        )

        if any(ctx.needs_input_grad[1:5]):
            op = lambda x: -1j * dt * (ham * x)
            lanczos_vectors_state, dS, lanczos_vectors_grad = double_krylov(
                op, state, grad_state_out, tolerance
            )
            # TODO: explore returning directly the basis in matrix form
            Vs = torch.stack(lanczos_vectors_state)
            del lanczos_vectors_state
            Vg = torch.stack(lanczos_vectors_grad)
            del lanczos_vectors_grad
            e_l = dS.mT @ Vs

        if ctx.needs_input_grad[1]:
            grad_omegas = torch.zeros_like(omegas)
            for i in range(nqubits):
                # dh as per the docstring
                dho = DHDOmegaSparse(i, e_l.device, nqubits, phis[i])
                # compute the trace
                v = dho @ e_l
                grad_omegas[i] = (-1j * dt * torch.tensordot(Vg.conj(), v)).real

        if ctx.needs_input_grad[2]:
            grad_deltas = torch.zeros_like(deltas)
            for i in range(nqubits):
                dhd = DHDDeltaSparse(i, nqubits)
                v = dhd @ e_l
                grad_deltas[i] = (-1j * dt * torch.tensordot(Vg.conj(), v)).real

        if ctx.needs_input_grad[3]:
            grad_phis = torch.zeros_like(phis)
            for i in range(nqubits):
                dhp = DHDPhiSparse(i, e_l.device, nqubits, omegas[i], phis[i])
                v = dhp @ e_l
                grad_phis[i] = (-1j * dt * torch.tensordot(Vg.conj(), v)).real

        if ctx.needs_input_grad[4]:
            grad_int_mat = torch.zeros_like(interaction_matrix)
            for i in range(nqubits):
                for j in range(i + 1, nqubits):
                    dhu = DHDUSparse(i, j, nqubits)
                    v = dhu @ e_l
                    grad_int_mat[i, j] = (-1j * dt * torch.tensordot(Vg.conj(), v)).real

        if ctx.needs_input_grad[5]:
            op = lambda x: (1j * dt) * (ham * x)
            grad_state_in = krylov_exp(op, grad_state_out.detach(), tolerance, tolerance)

        return (
            None,
            grad_omegas,
            grad_deltas,
            grad_phis,
            grad_int_mat,
            grad_state_in,
            None,
        )


class EvolveDensityMatrix:
    """Evolution of a density matrix under a Lindbladian operator."""

    @staticmethod
    def evolve(
        dt: float,
        omegas: torch.Tensor,
        deltas: torch.Tensor,
        phis: torch.Tensor,
        full_interaction_matrix: torch.Tensor,
        density_matrix: torch.Tensor,
        krylov_tolerance: float,
        pulser_lindblads: list[torch.Tensor],
    ) -> tuple[torch.Tensor, RydbergLindbladian]:
        ham = RydbergLindbladian(
            omegas=omegas,
            deltas=deltas,
            phis=phis,
            pulser_linblads=pulser_lindblads,
            interaction_matrix=full_interaction_matrix,
            device=density_matrix.device,
        )

        op = lambda x: -1j * dt * (ham @ x)
        return (
            krylov_exp(
                op,
                density_matrix,
                norm_tolerance=krylov_tolerance,
                exp_tolerance=krylov_tolerance,
                is_hermitian=False,
            ),
            ham,
        )



from pulser.backend.results import Results

from pulser.backend import (
    BitStrings,
    CorrelationMatrix,
    Energy,
    EnergyVariance,
    EnergySecondMoment,
    Expectation,
    Fidelity,
    Occupation,
    StateResult,
)

from .dense_operator import DenseOperator
from .sv_backend import SVBackend, SVConfig
from .state_vector import StateVector, inner
from .density_matrix_state import DensityMatrix


__all__ = [
    "__version__",
    "BitStrings",
    "CorrelationMatrix",
    "DenseOperator",
    "Energy",
    "EnergySecondMoment",
    "EnergyVariance",
    "Expectation",
    "Fidelity",
    "Occupation",
    "Results",
    "SVBackend",
    "SVConfig",
    "StateResult",
    "StateVector",
    "inner",
    "DensityMatrix",
]

__version__ = "2.4.0"
